function fem_sector_example(step)
%function fem_sector_example(step)
%
% Demonstration of FEM error convergence for sector example.
%
% This operation is very expensive (takes several hours) as hundred of
% thousands of point with global coordinates need to be searched and found
% in FEM meshes. This is required to be consistent in the error computation
% to other approximation techniques using these uniform test grids.
%
% The code can be run if RBmatlab from version 16.09 is installed
% as obtained from www.morepas.org/software 

% B. Haasdonk 8.5.2024

if nargin < 1
  step = 4;
end;

disp('FEM error convergence study for sector geometry')
disp('Note that the global to local coordinate search is extremely')
disp('expensive due to the many test points, and will overall take')
disp('several hours.')

switch step
  
%  case 1 %  create scale of refined gridfiles
%    fns = {'sectorg_alpha5over3','sectorg_alpha2over3'};
%    for fni = 1:length(fns)
%      fn = fns{fni}
%      [p,e,t] = initmesh(fn);
%      meshfile = [fn,'_r0.mat'];
%      save(meshfile,'p','e','t');
%      for i = 1:4
%	[p,e,t] = refinemesh(fn,p,e,t);
%	meshfile = [fn,'_r',num2str(i),'.mat'];
%	save(meshfile,'p','e','t');  
%      end;
%    end;
%    disp('mesh sequence generated and stored');

%    % continue with error computation: 
%    fem_sector_example(4);
    
 case 2 % load and plot mesh

  load('sectorg_alpha5over3_r1','p','t');
  grid = triagrid(p,t,[]);
  plot(grid);
  axis equal;

  figure;
  load('sectorg_alpha2over3_r1','p','t');
  grid = triagrid(p,t,[]);
  plot(grid);
  axis equal;

 case 3 % fem on mesh 
  
  params = [];
%  params.solution_number = 1; % example with smooth solution 
  params.solution_number = 2; % example with singular sol, inhom bv

%  params.alpha = 5/3;
%  params.grid_initfile = 'sectorg_alpha5over3_r1.mat';
%  params.grid_initfile = 'sectorg_alpha5over3_r2.mat';
%  params.grid_initfile = 'sectorg_alpha5over3_r3.mat';
%  params.grid_initfile = 'sectorg_alpha5over3_r4.mat';
  params.alpha = 2/3;
  params.grid_initfile = 'sectorg_alpha2over3_r1.mat';
%  params.grid_initfile = 'sectorg_alpha2over3_r2.mat';
%  params.grid_initfile = 'sectorg_alpha2over3_r3.mat';
%  params.grid_initfile = 'sectorg_alpha2over3_r4.mat';
  model = pacman_model(params);
  % generate grid and fem matrices:
  model_data = gen_model_data(model);
  figure, plot(model_data.grid);
  axis equal; axis tight; title('FEM grid')  
  sim_data = detailed_simulation(model, model_data); 
  % plot results
  figure, plot_sim_data(model,model_data,sim_data,[]);
  title('FEM solution of -Laplace u = f')
  axis equal;
  axis tight;
%  disp(['ndofs = ',num2str(sim_data.uh.df_info.ndofs)]);

 case 4 % fem on mesh and convergence study 
     
  params = [];
  params.solution_number = 1; % example with smooth solution 
%  params.solution_number = 2; % example with singular sol, inhom bv

    disp(' ');
    disp('ndofs  |    L2-error    |     H1-error    |     infty-error    |   L2-error-grid  | infty-error-grid    |    t_CPU')
    disp('---------------------------------------------------------------------------------------------------------------------')
    
  for i = 1:4
    % params.alpha = 5/3;
    % params.grid_initfile = ['sectorg_alpha5over3_r',num2str(i),'.mat'];
    %  params.grid_initfile = 'sectorg_alpha5over3_r2.mat';
    %  params.grid_initfile = 'sectorg_alpha5over3_r3.mat';
    %  params.grid_initfile = 'sectorg_alpha5over3_r4.mat';
    params.alpha = 2/3;
    params.grid_initfile = ['sectorg_alpha2over3_r',num2str(i),'.mat'];
    %  params.grid_initfile = 'sectorg_alpha2over3_r2.mat';
    %  params.grid_initfile = 'sectorg_alpha2over3_r3.mat';
    %  params.grid_initfile = 'sectorg_alpha2over3_r4.mat';
    model = pacman_model(params);
    % generate grid and fem matrices:
    tic
    model_data = gen_model_data(model);
    sim_data = detailed_simulation(model, model_data); 
    t = toc;
    % error computation:  
    % project exact solution onto higher degree polynomial fem func
    par.pdeg = 4;
    par.qdeg = 8;
    par.dimrange = 1;    
    p4_df_info = feminfo(par,model_data.grid);
    
    uexact_h = femdiscfunc([],p4_df_info);
    uexact_h = fem_interpol_global(model.solution,uexact_h);
    uh = femdiscfunc([],p4_df_info);
    u_local_eval = @(grid,elids,lcoord,params) ...
	my_uh_local_eval(grid,elids,lcoord,params,sim_data.uh); 
    uh = fem_interpol_local(u_local_eval,uh);
    
    err = uh - uexact_h;
    if i == 1
      plot(err);
      title('error for i=1')
      axis equal;
      axis tight;
    end;
%    keyboard;
    l2err = fem_l2_norm(err);
    h1err = fem_h1_norm(err);
    linftyerr = max(abs(err.dofs));
    [l2err_uniformgrid, linftyerr_uniformgrid] = errors_uniformgrid(model,uh);
%    l2err_uniformgrid = zeros(size(l2err));
%    linftyerr_uniformgrid = zeros(size(l2err));
    ndofs = sim_data.uh.df_info.ndofs;
    disp([num2str(ndofs,'%10.4d'),'   |   ',...
	  num2str(l2err,'%10.5e'),'   |   ',...
	  num2str(h1err,'%10.5e'),'   |   ',...
	  num2str(linftyerr,'%10.5e'),'   |   ',...
	  num2str(l2err_uniformgrid,'%10.5e'),'   |   ',...
	  num2str(linftyerr_uniformgrid,'%10.5e'),'   |   ',...
      num2str(t,'%10.5e')]);
  end;

 otherwise
  error('step number unknown');
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    auxiliary functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [l2err_uniformgrid, linftyerr_uniformgrid] = errors_uniformgrid(model, uh);  
% determine test points
  h = 0.002;
%  h = 0.002;
%  h = 0.002;
  x = -1:h:1;
  [XX,YY] = meshgrid(x,x);  
  i = find((XX.^2+YY.^2)<=1);
  XX = XX(i);
  YY = YY(i);
  Phi = atan(YY./XX);
  i = find(Phi<0 & XX>=0);
  Phi(i) = Phi(i) + 2*pi;
  i = find(XX<0);
  Phi(i) = Phi(i) + pi;
  i = find(Phi<model.alpha*pi);
  XX = XX(i);
  YY = YY(i);
  Phi = Phi(i);
  %scatter3(XX,YY,Phi)
  test_sol = model.solution([XX,YY]);
  test_uh = zeros(size(test_sol));  
  % evaluate fem approximation in global points
  % bad... : loop over points, should be vectorized...
  f = waitbar(0,'Iterating over points');
  for i = 1:length(Phi);
    if mod(i,1000)==0;
      waitbar(i/length(Phi),f,'Iterating over points');
    end;
    eind = find_triangle(uh.grid,[XX(i),YY(i)]);
    if eind>0
      lcoord = global2local(uh.grid,eind,[XX(i),YY(i)]);
      test_uh(i) = fem_evaluate(uh,eind,lcoord);
      % sanity check: transform local coordinate back to global:
      p = local2global(uh.grid,eind,lcoord,[]);
      d = p-[XX(i),YY(i)];
%      if norm(d)>10*eps
%	disp('norm of reconstructed point too large!, please inspect');%
%	keyboard
%      end;     
    else
      test_uh(i) = NaN;
    end;
  end;
  close(f);
  
  linftyerr_uniformgrid = max(abs(test_sol-test_uh));
  i = find(~isnan(test_uh));
  l2err_uniformgrid = sqrt(h*h*sum((test_sol(i)-test_uh(i)).^2));

  %figure;scatter3(XX,YY,double(isnan(test_uh)))
%  if isnan(l2err_uniformgrid)
%    disp('NaN in error!')
%    keyboard;
%  end;
  
function eind = find_triangle(grid,glob)
% returns the index of a triangle in grid containing the global point glob
% if no triangle is found, then -1 is returned.
inside = ones(grid.nelements,1);
for j = 1:3 % check if point is "above" edge connecting point j to j+1
  jp1 = mod(j,3)+1;
  jp2 = mod(jp1,3)+1;
  Xj = grid.X(grid.VI(:,j));
  Yj = grid.Y(grid.VI(:,j));
  Xjp1 = grid.X(grid.VI(:,jp1));
  Yjp1 = grid.Y(grid.VI(:,jp1));
  Vjjp1 = [Xjp1-Xj, Yjp1-Yj];
  Vjglob = [glob(1)*ones(size(Xj)) - Xj, glob(2)*ones(size(Yj)) - Yj];
  crossz = sign(Vjjp1(:,1).*Vjglob(:,2) - Vjjp1(:,2).*Vjglob(:,1));
  if j==1
    Xjp2 = grid.X(grid.VI(:,jp2));
    Yjp2 = grid.Y(grid.VI(:,jp2));
    Vjjp2 = [Xjp2-Xj, Yjp2-Yj];
    crossz2 = sign(Vjjp1(:,1).*Vjjp2(:,2) - Vjjp1(:,2).*Vjjp2(:,1)); % orientation of trias
  end;
  i = find(crossz.*crossz2<0);
  inside(i) = 0;
end;
eind = find(inside);
if isempty(eind)
  eind = -1;
end
if length(eind)>1
  %  disp('length eind > 1, please check! ');
  eind = eind(1);
end;
%if length(eind)==1
%  disp('length eind == 1, nice :-)  ');
%end;

% settings for pacman model
function model = pacman_model(params);
  if ~isfield(params,'alpha')
    alpha = 5/3;
  else
    alpha = params.alpha;
  end;
%  disp(['chosen alpha = ',num2str(alpha)]);
  %alpha = 5/3; % hard coded in sectorg.m if changed here, change there!
  model = poisson_model(params);
  model.alpha = alpha;
  model = rmfield(model,{'boundary_type','normals',...
			 'xnumintervals','ynumintervals','xrange','yrange'});
  model.has_reaction = 0;
  model.has_advection = 0;
  model.has_diffusivity = 1;
  model.has_source = 1;
  model.has_dirichlet_values = 1;
  model.has_neumann_values = 0;
  model.has_robin_values = 0;
  model.compute_output_functional = 0;
  switch params.solution_number
    case 1 % smooth solution
      model.solution = @(glob,params) ...
			sum(glob.^2,2);
      model.source = @(glob,params) ...
		     - 4 * ones(size(glob,1),1);
    case 2 % solution with singularity and inhomogeneous bnd val.
      model.solution = @(glob,params) ...
			pacman_exact_solution(glob',alpha);
      model.source = @(glob,params) ...
		      neg_Laplace_pacman_exact_solution(glob',alpha);
      % case 3 % solution wiht singularity and homogeneous bnd val.
      %  % someting seems to be buggy here, no convergence observed...
      %  model.solution = @(glob,params) ...
      %      pacman_exact_solution2(glob',alpha);
      %  model.source = @(glob,params) ...
      %      neg_Laplace_pacman_exact_solution2(glob',alpha);
      %  error('please use solution_number=1 or 2.')
  end;
  model.diffusivity_tensor = @(glob,params) ...
			      [ones(size(glob,1),1),...
			       zeros(size(glob,1),1),...
			       zeros(size(glob,1),1),...
			       ones(size(glob,1),1)];
  model.reaction = @(glob,params) zeros(size(glob,1),1);
  model.dirichlet_values = @(glob,params) ...
			    params.solution(glob,params);
  model.grid_initfile = params.grid_initfile;
  model.gridtype = 'triagrid';
  model.pdeg = 1;
  model.qdeg = 2;
  model.dimrange = 1;
  model = elliptic_discrete_model(model);
	     %model.detailed_simulation = @pacman_detailed_simulation;

function f = pacman_exact_solution(x,alpha);
% function u(x) = |x|^(1/alpha)sin(phi(x)/alpha)
% with phi(x) = atan(x2/x1).
% which has -Laplace u = 0 on pacman shape
% but non-homogeneous boundary values.
f1 = sum(x.^2,1).^(0.5/alpha);
phi = atan(x(2,:)./x(1,:));
i = find(phi<0 & x(1,:)>=0);
phi(i) = phi(i) + 2*pi;
i = find(x(1,:)<0);
phi(i) = phi(i) + pi;
f2 = sin(phi/alpha);
i = find(isnan(f2));
f2(i) = 0;
f = f1.*f2;

function f = neg_Laplace_pacman_exact_solution(x,alpha);
f = zeros(size(x,2),1);

function res = my_uh_local_eval(grid,elids,lcoord,params,df)
% dummy function used for evaluating a discrete function at finer
% lagrange-grid nodes
res = fem_evaluate(df,elids,lcoord,[],[]);
